      Program Appendix_7_3
C
C
C  This is a main program for the example in section 7.3.
C  This example, in parallel, solves a system of linear equations in
C  the form [A]{X}={B} by the asynchronous Gauss-Seidal Iteration.
C  
C  global variables      
C
      INTEGER*4 Order
      PARAMETER (Order=200)
      INTEGER*4 NumberOfTasks
      PARAMETER (NumberOfTasks=4)
      INTEGER*4 GrainSize
      PARAMETER (GrainSize=1000)
      INTEGER*4 Partitions
      PARAMETER (Partitions=8)
      INTEGER*4 N,LockID,Event(2),Subset,Locks(Partitions)
      INTEGER*4 Outstanding,Tasks,First(Partitions+1)
      INTEGER*4 J4TEMP,K4TEMP,SEED,TaskID(NumberOfTasks)
      INTEGER*4 TaskIndex(NumberOfTasks),ActiveEvent
      REAL*4 A(Order,Order),X(Order),B(Order)
      REAL*4 WORK(NumberOfTasks),Residual(NumberOfTasks),Tolerance
      LOGICAL*4 Done
C
C  declare dispatched subroutine
C
      EXTERNAL SubA
      DATA Tolerance/0.001/,SEED/88888888/
      DATA First/1,26,51,76,101,126,151,176,201/
C
C  generate data in a sequential procedure
C
      WRITE(*,'('' >>> Generate data ...'')')
      Tasks=NumberOfTasks
      N=Order
      DO J4TEMP=1,N*N
         A(J4TEMP,1)=RAN(SEED)
      END DO
      DO J4TEMP=1,N
         DO K4TEMP=1,N
            A(J4TEMP,J4TEMP)=A(J4TEMP,J4TEMP)+A(J4TEMP,K4TEMP)
         END DO
         A(J4TEMP,J4TEMP)=A(J4TEMP,J4TEMP)*5.0
      END DO
      DO J4TEMP=1,N
         B(J4TEMP)=0.0
         DO K4TEMP=1,N
            B(J4TEMP)=B(J4TEMP)+A(J4TEMP,K4TEMP)
         END DO
      END DO
      DO J4TEMP=1,N
         X(J4TEMP)=0.0
      END DO
C
C  create parallel locks
C
      CALL CreateParallelLock(LockID)
      DO J4TEMP=1,Partitions
         CALL CreateParallelLock(Locks(J4TEMP))
      END DO
C
C  create parallel events
C
      CALL CreateParallelEvent(Event(1))
      CALL CreateParallelEvent(Event(2))
      ActiveEvent=1
C
C  determine length of subsets
C
      Subset=MIN0(GrainSize/(2*N+5),Order/NumberOfTasks)
      IF(Subset.EQ.0) Subset=1
C
C  initialize
C
      Outstanding=1
      CALL InitializeParallelEvent(Event(1),Tasks)
C
C  solve in parallel
C
      WRITE(*,'('' >>> Solve in parallel ...'')')
      DO J4TEMP=1,NumberOfTasks-1
         TaskIndex(J4TEMP)=J4TEMP
         CALL Dispatch_18(TaskID(J4TEMP),SubA,A,X,B,N,
     &                    LockID,Subset,Outstanding,
     &                    Partitions,First,Locks,
     &                    TaskIndex(J4TEMP),Tasks,
     &                    Work,Residual,
     &                    Tolerance,Done,
     &                    Event,ActiveEvent)
      END DO
      CALL SubA(A,X,B,N,
     &          LockID,Subset,Outstanding,
     &          Partitions,First,Locks,
     &          NumberOfTasks,Tasks,
     &          Work,Residual,
     &          Tolerance,Done,
     &          Event,ActiveEvent)
C
C  terminate tasks
C
      CALL WaitForAndTerminateAllTasks
C
C  delete parallel lock
C
      CALL DeleteParallelLock(LockID)
      DO J4TEMP=1,Partitions
         CALL DeleteParallelLock(Locks(J4TEMP))
      END DO
C
C  delete parallel events
C
      CALL DeleteParallelEvent(Event(1))
      CALL DeleteParallelEvent(Event(2))
C
C  output
C
      WRITE(*,'(''      X(  1)= '',F7.4)') X(  1)
      WRITE(*,'(''      X( 50)= '',F7.4)') X( 50)
      WRITE(*,'(''      X(100)= '',F7.4)') X(100)
      WRITE(*,'(''      X(150)= '',F7.4)') X(150)
      WRITE(*,'(''      X(200)= '',F7.4)') X(200)
C
      STOP
      END
      SUBROUTINE SubA(A,X,B,N,
     &                LockID,Subset,Outstanding,
     &                Partitions,First,Locks,
     &                TaskIndex,Tasks,
     &                Work,Residual,
     &                Tolerance,Done,
     &                Event,ActiveEvent)
C
C
C  routine to solve a system of liner equations by asynchronous
C  Gauss-Seidal iteration
C  (A)FORTRAN CALL: CALL SubA(A,X,B,N,
C                             LockID,Subset,Outstanding,
C                             Partitions,First,Locks,
C                             TaskIndex,Tasks,
C                             Work,Residual,
C                             Tolerance,Done,
C                             Event,ActiveEvent)
C     1.A: <R4> left side positive definite matrix, dimension(N,N)
C     2.X: <R4> solution to be determined, dimension(N)
C     3.B: <R4> right side vector, dimension(N)
C     4.N: <I4> order
C     5.LockID: <I4> identifier of lock that protects distribution of
C                    subsets
C     6.Subset: <I4> length of subset
C     7.Outstanding: <I4> outstanding element that must be initialized
C                         to one
C     8.Partitions: <I4> number of critical subsections
C     9.First: <I4> the first index in each critical subsection,
C                   dimension(Partitions+1),
C                   in which the last element = N+1
C    10.Locks: <I4> identifier of locks for each critical subsection,
C                   dimension(Partitions)
C    11.TaskIndex: <I4> task ordinal number
C    12.Tasks: <I4> number of tasks
C    13.Work: <R4> working array for computing an increment,
C                  dimension(Tasks)
C    14.Residual: <R4> working array for the maximal residual in
C                      each task, dimension(Tasks)
C    15.Tolerance: <R4> tolerance of residual
C    16.Done: <L4> working variable to indicate if the job has been
C                  done
C    17.Event: <I4> identifiers of parallel event that handles task
C                   synchronization, dimension(2)
C                   The first parallel event Event(1) must be
C                   initialized in the caller.
C    18.ActiveEvent: <I4> the active parallel event which must be
C                         initialized to one in the caller
C
C  dummy arguments
C
      INTEGER*4 Partitions,Subset,Outstanding,LockID,First(1)
      INTEGER*4 TaskIndex,Tasks,Event(2),ActiveEvent,N,Locks(1)
      REAL*4 A(N,1),X(1),B(1),Work(1),Residual(1),Tolerance
      LOGICAL*4 Done
C
C  private varibales
C
      INTEGER*4 J4TEMP$,K4TEMP$,L4TEMP$,M4TEMP$,N4TEMP$
      LOGICAL*4 YES$
      AUTOMATIC J4TEMP$,K4TEMP$,L4TEMP$,M4TEMP$,N4TEMP$,YES$
C
C  entry point to a loop
C
      Done=.False.
10    Residual(TaskIndex)=0.0
      L4TEMP$=1
C
C  [critical section]
C  distribute subsets
C
20    CALL LockonParallelLock(LockID)
      IF(Outstanding.GT.N) THEN
         CALL UnlockParallelLock(LockID)
         GO TO 30
      END IF
      J4TEMP$=Outstanding
      Outstanding=Outstanding+Subset
      CALL UnlockParallelLock(LockID)
C
C  [parallelism]
C  set initial value for the increment
C  total size = 2*N+5
C
      DO K4TEMP$=J4TEMP$,MIN0(J4TEMP$+Subset-1,N)
         WORK(TaskIndex)=B(K4TEMP$)
C
C  [implicit parallelism]
C  compute in each critical subsection
C
         DO M4TEMP$=1,Partitions
            CALL LockonParallelLock(Locks(M4TEMP$))
            DO N4TEMP$=First(M4TEMP$),First(M4TEMP$+1)-1
               WORK(TaskIndex)=WORK(TaskIndex)-A(K4TEMP$,N4TEMP$)*
     &                                         X(N4TEMP$)
            END DO
            CALL UnlockParallelLock(Locks(M4TEMP$))
         END DO
C
C  [parallelism]
C  compute an increment
C
         WORK(TaskIndex)=WORK(TaskIndex)/A(K4TEMP$,K4TEMP$)
         IF(ABS(WORK(TaskIndex)).GT.Residual(TaskIndex)) THEN
            Residual(TaskIndex)=ABS(WORK(TaskIndex))
         END IF
C
C  [implicit parallelism]
C  update an element of {X}
C
         DO WHILE (First(L4TEMP$+1).LE.K4TEMP$)
            L4TEMP$=L4TEMP$+1
         END DO
         CALL LockonParallelLock(Locks(L4TEMP$))
         X(K4TEMP$)=X(K4TEMP$)+WORK(TaskIndex)
         CALL UnlockParallelLock(Locks(L4TEMP$))
      END DO         
      GO TO 20
C
C  [task synchronization]
C  complete a cycle of parallel event
C
30    CALL PostParallelEvent(Event(ActiveEvent))
      CALL IfCompleteParallelEvent(Event(ActiveEvent),YES$)
C
C  [subsequently sequential fragment]
C  check accuracy, and initialize outstanding element and
C  another parallel event
C
      IF(.NOT.YES$) THEN
         K4TEMP$=0
         DO J4TEMP$=1,Tasks
            IF(Residual(J4TEMP$).GT.Tolerance) THEN
               Outstanding=1
               SELECT CASE (ActiveEvent)
               CASE (1)
                  ActiveEvent=2
                  K4TEMP$=1
               CASE (2)
                  ActiveEvent=1
                  K4TEMP$=2
               END SELECT
               CALL InitializeParallelEvent(Event(ActiveEvent),Tasks)
               EXIT
            END IF
         END DO
         IF(K4TEMP$.EQ.0) THEN
            Done=.True.
            CALL CompleteParallelEvent(Event(ActiveEvent))
         ELSE
            CALL CompleteParallelEvent(Event(K4TEMP$))
         END IF
      END IF
C
C  return for convergence of solution
C
      IF(DONE) RETURN
C
C  go to one more loop
C
      GO TO 10
      END
      REAL*4 FUNCTION RAN(SEED)
C
C
C  function to generate a pseudo-random number
C  (A)FORTRAN ENTRY: RAN(SEED)
C     1.RAN: <R4> return a random number
C     2.SEED: <I4> a seed (updated)
C
      INTEGER*4 SEED,A
      REAL*8 RTEMPA,RTEMPB
      DATA M/2147483647/,A/16807/
C
C  generate a random number
C
      RTEMPA=DFLOAT(SEED)*A
      ITEMP=INT(RTEMPA/DFLOAT(M))
      RTEMPB=DFLOAT(M)*DFLOAT(ITEMP)
      SEED=INT(RTEMPA-RTEMPB)
      RAN=DFLOAT(SEED)/DFLOAT(M)
C
      RETURN
      END
 
